@@ -3,12 +3,12 @@
  * syscache.c
  *	  System cache management routines
  *
- * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2010, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/utils/cache/syscache.c,v 1.108 2006/10/06 18:23:35 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/cache/syscache.c,v 1.126 2010/02/14 18:42:17 rhaas Exp $
  *
  * NOTES
  *	  These routines allow the parser/planner/executor to perform
@@ -20,7 +20,7 @@
  */
 #include "postgres.h"
 
-#include "access/heapam.h"
+#include "access/sysattr.h"
 #include "catalog/indexing.h"
 #include "catalog/pg_aggregate.h"
 #include "catalog/pg_amop.h"
@@ -28,17 +28,30 @@
 #include "catalog/pg_auth_members.h"
 #include "catalog/pg_authid.h"
 #include "catalog/pg_cast.h"
+#include "catalog/pg_constraint.h"
 #include "catalog/pg_conversion.h"
 #include "catalog/pg_database.h"
-#include "catalog/pg_inherits.h"
+#include "catalog/pg_default_acl.h"
+#include "catalog/pg_enum.h"
+#include "catalog/pg_foreign_data_wrapper.h"
+#include "catalog/pg_foreign_server.h"
 #include "catalog/pg_language.h"
 #include "catalog/pg_namespace.h"
 #include "catalog/pg_opclass.h"
 #include "catalog/pg_operator.h"
+#include "catalog/pg_opfamily.h"
 #include "catalog/pg_proc.h"
 #include "catalog/pg_rewrite.h"
 #include "catalog/pg_statistic.h"
+#include "catalog/pg_tablespace.h"
+#include "catalog/pg_ts_config.h"
+#include "catalog/pg_ts_config_map.h"
+#include "catalog/pg_ts_dict.h"
+#include "catalog/pg_ts_parser.h"
+#include "catalog/pg_ts_template.h"
 #include "catalog/pg_type.h"
+#include "catalog/pg_user_mapping.h"
+#include "utils/rel.h"
 #include "utils/syscache.h"
 
 
@@ -46,18 +59,13 @@
 
 	Adding system caches:
 
-	Add your new cache to the list in include/utils/syscache.h.  Keep
-	the list sorted alphabetically and adjust the cache numbers
-	accordingly.
+	Add your new cache to the list in include/utils/syscache.h.
+	Keep the list sorted alphabetically.
 
 	Add your entry to the cacheinfo[] array below. All cache lists are
 	alphabetical, so add it in the proper place.  Specify the relation OID,
-	index OID, number of keys, key attribute numbers, and number of hash
-	buckets.  If the relation contains tuples that are associated with a
-	particular relation (for example, its attributes, rules, triggers, etc)
-	then specify the attribute number that contains the OID of the associated
-	relation.  This is used by CatalogCacheFlushRelation() to remove the
-	correct tuples during a table drop or relcache invalidation event.
+	index OID, number of keys, key attribute numbers, and initial number of
+	hash buckets.
 
 	The number of hash buckets must be a power of 2.  It's reasonable to
 	set this to the number of entries that might be in the particular cache
@@ -86,7 +94,6 @@ struct cachedesc
 {
 	Oid			reloid;			/* OID of the relation being cached */
 	Oid			indoid;			/* OID of index relation for this cache */
-	int			reloidattr;		/* attr number of rel OID reference, or 0 */
 	int			nkeys;			/* # of keys needed for cache lookup */
 	int			key[4];			/* attribute numbers of key attrs */
 	int			nbuckets;		/* number of hash buckets for this cache */
@@ -95,7 +102,6 @@ struct cachedesc
 static const struct cachedesc cacheinfo[] = {
 	{AggregateRelationId,		/* AGGFNOID */
 		AggregateFnoidIndexId,
-		0,
 		1,
 		{
 			Anum_pg_aggregate_aggfnoid,
@@ -107,7 +113,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AccessMethodRelationId,	/* AMNAME */
 		AmNameIndexId,
-		0,
 		1,
 		{
 			Anum_pg_am_amname,
@@ -119,7 +124,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AccessMethodRelationId,	/* AMOID */
 		AmOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -131,11 +135,10 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AccessMethodOperatorRelationId,	/* AMOPOPID */
 		AccessMethodOperatorIndexId,
-		0,
 		2,
 		{
 			Anum_pg_amop_amopopr,
-			Anum_pg_amop_amopclaid,
+			Anum_pg_amop_amopfamily,
 			0,
 			0
 		},
@@ -143,31 +146,28 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AccessMethodOperatorRelationId,	/* AMOPSTRATEGY */
 		AccessMethodStrategyIndexId,
-		0,
-		3,
+		4,
 		{
-			Anum_pg_amop_amopclaid,
-			Anum_pg_amop_amopsubtype,
-			Anum_pg_amop_amopstrategy,
-			0
+			Anum_pg_amop_amopfamily,
+			Anum_pg_amop_amoplefttype,
+			Anum_pg_amop_amoprighttype,
+			Anum_pg_amop_amopstrategy
 		},
 		64
 	},
 	{AccessMethodProcedureRelationId,	/* AMPROCNUM */
 		AccessMethodProcedureIndexId,
-		0,
-		3,
+		4,
 		{
-			Anum_pg_amproc_amopclaid,
-			Anum_pg_amproc_amprocsubtype,
-			Anum_pg_amproc_amprocnum,
-			0
+			Anum_pg_amproc_amprocfamily,
+			Anum_pg_amproc_amproclefttype,
+			Anum_pg_amproc_amprocrighttype,
+			Anum_pg_amproc_amprocnum
 		},
 		64
 	},
 	{AttributeRelationId,		/* ATTNAME */
 		AttributeRelidNameIndexId,
-		Anum_pg_attribute_attrelid,
 		2,
 		{
 			Anum_pg_attribute_attrelid,
@@ -179,7 +179,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AttributeRelationId,		/* ATTNUM */
 		AttributeRelidNumIndexId,
-		Anum_pg_attribute_attrelid,
 		2,
 		{
 			Anum_pg_attribute_attrelid,
@@ -191,7 +190,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AuthMemRelationId,			/* AUTHMEMMEMROLE */
 		AuthMemMemRoleIndexId,
-		0,
 		2,
 		{
 			Anum_pg_auth_members_member,
@@ -203,7 +201,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AuthMemRelationId,			/* AUTHMEMROLEMEM */
 		AuthMemRoleMemIndexId,
-		0,
 		2,
 		{
 			Anum_pg_auth_members_roleid,
@@ -215,7 +212,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AuthIdRelationId,			/* AUTHNAME */
 		AuthIdRolnameIndexId,
-		0,
 		1,
 		{
 			Anum_pg_authid_rolname,
@@ -227,7 +223,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{AuthIdRelationId,			/* AUTHOID */
 		AuthIdOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -240,7 +235,6 @@ static const struct cachedesc cacheinfo[] = {
 	{
 		CastRelationId,			/* CASTSOURCETARGET */
 		CastSourceTargetIndexId,
-		0,
 		2,
 		{
 			Anum_pg_cast_castsource,
@@ -252,10 +246,9 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{OperatorClassRelationId,	/* CLAAMNAMENSP */
 		OpclassAmNameNspIndexId,
-		0,
 		3,
 		{
-			Anum_pg_opclass_opcamid,
+			Anum_pg_opclass_opcmethod,
 			Anum_pg_opclass_opcname,
 			Anum_pg_opclass_opcnamespace,
 			0
@@ -264,7 +257,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{OperatorClassRelationId,	/* CLAOID */
 		OpclassOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -276,7 +268,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{ConversionRelationId,		/* CONDEFAULT */
 		ConversionDefaultIndexId,
-		0,
 		4,
 		{
 			Anum_pg_conversion_connamespace,
@@ -288,7 +279,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{ConversionRelationId,		/* CONNAMENSP */
 		ConversionNameNspIndexId,
-		0,
 		2,
 		{
 			Anum_pg_conversion_conname,
@@ -298,9 +288,19 @@ static const struct cachedesc cacheinfo[] = {
 		},
 		128
 	},
-	{ConversionRelationId,		/* CONOID */
+	{ConstraintRelationId,		/* CONSTROID */
+		ConstraintOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		1024
+	},
+	{ConversionRelationId,		/* CONVOID */
 		ConversionOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -312,7 +312,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{DatabaseRelationId,		/* DATABASEOID */
 		DatabaseOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -322,9 +321,85 @@ static const struct cachedesc cacheinfo[] = {
 		},
 		4
 	},
+	{DefaultAclRelationId,		/* DEFACLROLENSPOBJ */
+		DefaultAclRoleNspObjIndexId,
+		3,
+		{
+			Anum_pg_default_acl_defaclrole,
+			Anum_pg_default_acl_defaclnamespace,
+			Anum_pg_default_acl_defaclobjtype,
+			0
+		},
+		256
+	},
+	{EnumRelationId,			/* ENUMOID */
+		EnumOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		256
+	},
+	{EnumRelationId,			/* ENUMTYPOIDNAME */
+		EnumTypIdLabelIndexId,
+		2,
+		{
+			Anum_pg_enum_enumtypid,
+			Anum_pg_enum_enumlabel,
+			0,
+			0
+		},
+		256
+	},
+	{ForeignDataWrapperRelationId,		/* FOREIGNDATAWRAPPERNAME */
+		ForeignDataWrapperNameIndexId,
+		1,
+		{
+			Anum_pg_foreign_data_wrapper_fdwname,
+			0,
+			0,
+			0
+		},
+		8
+	},
+	{ForeignDataWrapperRelationId,		/* FOREIGNDATAWRAPPEROID */
+		ForeignDataWrapperOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		8
+	},
+	{ForeignServerRelationId,	/* FOREIGNSERVERNAME */
+		ForeignServerNameIndexId,
+		1,
+		{
+			Anum_pg_foreign_server_srvname,
+			0,
+			0,
+			0
+		},
+		32
+	},
+	{ForeignServerRelationId,	/* FOREIGNSERVEROID */
+		ForeignServerOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		32
+	},
 	{IndexRelationId,			/* INDEXRELID */
 		IndexRelidIndexId,
-		Anum_pg_index_indrelid,
 		1,
 		{
 			Anum_pg_index_indexrelid,
@@ -334,21 +409,8 @@ static const struct cachedesc cacheinfo[] = {
 		},
 		1024
 	},
-	{InheritsRelationId,		/* INHRELID */
-		InheritsRelidSeqnoIndexId,
-		Anum_pg_inherits_inhrelid,
-		2,
-		{
-			Anum_pg_inherits_inhrelid,
-			Anum_pg_inherits_inhseqno,
-			0,
-			0
-		},
-		256
-	},
 	{LanguageRelationId,		/* LANGNAME */
 		LanguageNameIndexId,
-		0,
 		1,
 		{
 			Anum_pg_language_lanname,
@@ -360,7 +422,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{LanguageRelationId,		/* LANGOID */
 		LanguageOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -372,7 +433,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{NamespaceRelationId,		/* NAMESPACENAME */
 		NamespaceNameIndexId,
-		0,
 		1,
 		{
 			Anum_pg_namespace_nspname,
@@ -384,7 +444,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{NamespaceRelationId,		/* NAMESPACEOID */
 		NamespaceOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -396,7 +455,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{OperatorRelationId,		/* OPERNAMENSP */
 		OperatorNameNspIndexId,
-		0,
 		4,
 		{
 			Anum_pg_operator_oprname,
@@ -408,7 +466,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{OperatorRelationId,		/* OPEROID */
 		OperatorOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -418,9 +475,30 @@ static const struct cachedesc cacheinfo[] = {
 		},
 		1024
 	},
+	{OperatorFamilyRelationId,	/* OPFAMILYAMNAMENSP */
+		OpfamilyAmNameNspIndexId,
+		3,
+		{
+			Anum_pg_opfamily_opfmethod,
+			Anum_pg_opfamily_opfname,
+			Anum_pg_opfamily_opfnamespace,
+			0
+		},
+		64
+	},
+	{OperatorFamilyRelationId,	/* OPFAMILYOID */
+		OpfamilyOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		64
+	},
 	{ProcedureRelationId,		/* PROCNAMEARGSNSP */
 		ProcedureNameArgsNspIndexId,
-		0,
 		3,
 		{
 			Anum_pg_proc_proname,
@@ -432,7 +510,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{ProcedureRelationId,		/* PROCOID */
 		ProcedureOidIndexId,
-		0,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -444,7 +521,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{RelationRelationId,		/* RELNAMENSP */
 		ClassNameNspIndexId,
-		ObjectIdAttributeNumber,
 		2,
 		{
 			Anum_pg_class_relname,
@@ -456,7 +532,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{RelationRelationId,		/* RELOID */
 		ClassOidIndexId,
-		ObjectIdAttributeNumber,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -468,7 +543,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{RewriteRelationId,			/* RULERELNAME */
 		RewriteRelRulenameIndexId,
-		Anum_pg_rewrite_ev_class,
 		2,
 		{
 			Anum_pg_rewrite_ev_class,
@@ -478,21 +552,129 @@ static const struct cachedesc cacheinfo[] = {
 		},
 		1024
 	},
-	{StatisticRelationId,		/* STATRELATT */
-		StatisticRelidAttnumIndexId,
-		Anum_pg_statistic_starelid,
-		2,
+	{StatisticRelationId,		/* STATRELATTINH */
+		StatisticRelidAttnumInhIndexId,
+		3,
 		{
 			Anum_pg_statistic_starelid,
 			Anum_pg_statistic_staattnum,
-			0,
+			Anum_pg_statistic_stainherit,
 			0
 		},
 		1024
 	},
+	{TableSpaceRelationId,		/* TABLESPACEOID */
+		TablespaceOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0,
+		},
+		16
+	},
+	{TSConfigMapRelationId,		/* TSCONFIGMAP */
+		TSConfigMapIndexId,
+		3,
+		{
+			Anum_pg_ts_config_map_mapcfg,
+			Anum_pg_ts_config_map_maptokentype,
+			Anum_pg_ts_config_map_mapseqno,
+			0
+		},
+		4
+	},
+	{TSConfigRelationId,		/* TSCONFIGNAMENSP */
+		TSConfigNameNspIndexId,
+		2,
+		{
+			Anum_pg_ts_config_cfgname,
+			Anum_pg_ts_config_cfgnamespace,
+			0,
+			0
+		},
+		16
+	},
+	{TSConfigRelationId,		/* TSCONFIGOID */
+		TSConfigOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		16
+	},
+	{TSDictionaryRelationId,	/* TSDICTNAMENSP */
+		TSDictionaryNameNspIndexId,
+		2,
+		{
+			Anum_pg_ts_dict_dictname,
+			Anum_pg_ts_dict_dictnamespace,
+			0,
+			0
+		},
+		16
+	},
+	{TSDictionaryRelationId,	/* TSDICTOID */
+		TSDictionaryOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		16
+	},
+	{TSParserRelationId,		/* TSPARSERNAMENSP */
+		TSParserNameNspIndexId,
+		2,
+		{
+			Anum_pg_ts_parser_prsname,
+			Anum_pg_ts_parser_prsnamespace,
+			0,
+			0
+		},
+		4
+	},
+	{TSParserRelationId,		/* TSPARSEROID */
+		TSParserOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		4
+	},
+	{TSTemplateRelationId,		/* TSTEMPLATENAMENSP */
+		TSTemplateNameNspIndexId,
+		2,
+		{
+			Anum_pg_ts_template_tmplname,
+			Anum_pg_ts_template_tmplnamespace,
+			0,
+			0
+		},
+		16
+	},
+	{TSTemplateRelationId,		/* TSTEMPLATEOID */
+		TSTemplateOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		16
+	},
 	{TypeRelationId,			/* TYPENAMENSP */
 		TypeNameNspIndexId,
-		Anum_pg_type_typrelid,
 		2,
 		{
 			Anum_pg_type_typname,
@@ -504,7 +686,6 @@ static const struct cachedesc cacheinfo[] = {
 	},
 	{TypeRelationId,			/* TYPEOID */
 		TypeOidIndexId,
-		Anum_pg_type_typrelid,
 		1,
 		{
 			ObjectIdAttributeNumber,
@@ -513,6 +694,28 @@ static const struct cachedesc cacheinfo[] = {
 			0
 		},
 		1024
+	},
+	{UserMappingRelationId,		/* USERMAPPINGOID */
+		UserMappingOidIndexId,
+		1,
+		{
+			ObjectIdAttributeNumber,
+			0,
+			0,
+			0
+		},
+		128
+	},
+	{UserMappingRelationId,		/* USERMAPPINGUSERSERVER */
+		UserMappingUserServerIndexId,
+		2,
+		{
+			Anum_pg_user_mapping_umuser,
+			Anum_pg_user_mapping_umserver,
+			0,
+			0
+		},
+		128
 	}
 };
 
@@ -526,7 +729,7 @@ static bool CacheInitialized = false;
  * InitCatalogCache - initialize the caches
  *
  * Note that no database access is done here; we only allocate memory
- * and initialize the cache structure.	Interrogation of the database
+ * and initialize the cache structure.  Interrogation of the database
  * to complete initialization of a cache happens upon first use
  * of that cache.
  */
@@ -544,7 +747,6 @@ InitCatalogCache(void)
 		SysCache[cacheId] = InitCatCache(cacheId,
 										 cacheinfo[cacheId].reloid,
 										 cacheinfo[cacheId].indoid,
-										 cacheinfo[cacheId].reloidattr,
 										 cacheinfo[cacheId].nkeys,
 										 cacheinfo[cacheId].key,
 										 cacheinfo[cacheId].nbuckets);
@@ -707,10 +909,9 @@ SearchSysCacheAttName(Oid relid, const char *attname)
 {
 	HeapTuple	tuple;
 
-	tuple = SearchSysCache(ATTNAME,
-						   ObjectIdGetDatum(relid),
-						   CStringGetDatum(attname),
-						   0, 0);
+	tuple = SearchSysCache2(ATTNAME,
+							ObjectIdGetDatum(relid),
+							CStringGetDatum(attname));
 	if (!HeapTupleIsValid(tuple))
 		return NULL;
 	if (((Form_pg_attribute) GETSTRUCT(tuple))->attisdropped)
@@ -765,7 +966,7 @@ SearchSysCacheExistsAttName(Oid relid, const char *attname)
  *		extract a specific attribute.
  *
  * This is equivalent to using heap_getattr() on a tuple fetched
- * from a non-cached relation.	Usually, this is only used for attributes
+ * from a non-cached relation.  Usually, this is only used for attributes
  * that could be NULL or variable length; the fixed-size attributes in
  * a system table are accessed just by mapping the tuple onto the C struct
  * declarations from include/catalog/.
