@@ -1,9 +1,9 @@
 /*
  * psql - the PostgreSQL interactive terminal
  *
- * Copyright (c) 2000-2003, PostgreSQL Global Development Group
+ * Copyright (c) 2000-2010, PostgreSQL Global Development Group
  *
- * $Header: /cvsroot/pgsql/src/bin/psql/prompt.h,v 1.12 2003/08/04 23:59:40 tgl Exp $
+ * $PostgreSQL: pgsql/src/bin/psql/prompt.h,v 1.21 2010/01/02 16:57:59 momjian Exp $
  */
 #ifndef PROMPT_H
 #define PROMPT_H
@@ -15,6 +15,7 @@ typedef enum _promptStatus
 	PROMPT_COMMENT,
 	PROMPT_SINGLEQUOTE,
 	PROMPT_DOUBLEQUOTE,
+	PROMPT_DOLLARQUOTE,
 	PROMPT_PAREN,
 	PROMPT_COPY
 } promptStatus_t;
